% this program performs Motion compensation 
% input: 
%    video in YUV 420 format
%    
%    
%    
%    
% output:
%    prediction of 
%    
%   Copyright 2014 Aline ROUMY, Inria.
%% Parameters
video='mobile';
NbCols = 352; NbRows = 288; % CIF
FrameRate = 30; NbFrame = 50;

video='BasketBallPass';
NbCols = 416; NbRows = 240; 
FrameRate = 50; NbFrame = 150;

video='akiyo';
NbCols = 176; NbRows = 144; 
FrameRate = 50; NbFrame = 100;

video='foreman';
NbCols = 352; NbRows = 288; % CIF
FrameRate = 30; NbFrame = 50;

referenceFrameNumber = 1; % number of the frame taken as a reference
currentFrameNumber = 20; % number of the frame to be processed

s=8; % size of the search window = [-s ,s]
nbRowBlock=32;nbColBlock=32; % size of the block for Block matching motion estimation

%% Read a movie file, extract 2 images (reference,current) and display the Y compoennt of the 2 images

videoFilename=sprintf('%s_%dx%d_%d_%d.yuv',video,NbCols,NbRows,FrameRate,NbFrame);

referenceFrame = yuv2frame(videoFilename,NbCols,NbRows,referenceFrameNumber);
currentFrame = yuv2frame(videoFilename,NbCols,NbRows,currentFrameNumber);

referenceImage= referenceFrame(:, :, 1);
currentImage= currentFrame(:, :, 1);
  
figure(1)
image(referenceImage); colormap(gray(256)); axis image; axis off; title(sprintf('Y component reference image, Frame: %d',referenceFrameNumber));
figure(2)
image(currentImage); colormap(gray(256)); axis image; axis off; title(sprintf('Y component current image, Frame: %d',currentFrameNumber));
figure(3)
image(128+currentImage-referenceImage); colormap(gray(256)); axis image; axis off; title('difference image');
% + 128 because the difference ~= [-128, 127]
% avoid imagesc that rescales the image to be drawn

%% Motion Vector estimation
% full search, non regularized

[nbRow,nbCol] = size(referenceImage);
predictedImage = zeros(nbRow,nbCol);
MotionVector=[];
MotionVectorPosition=[];

%fprintf('%d %d \n',nbRow,nbCol);
tic
for ir=1:nbRowBlock:nbRow,
    for ic=1:nbColBlock:nbCol,
        % size of the Block s.t. it fits within the frame 
        nbRowBlockTested=min(nbRowBlock,nbRow-ir+1); % could have been computed outside this loop. Put here for readibility of the code
        nbColBlockTested=min(nbColBlock,nbCol-ic+1); 
        %fprintf('Current block size: %d %d \n',nbRowBlockTested,nbColBlockTested);    % Debug
        
        % current block selection
        currentBlock=currentImage(ir:ir+nbRowBlockTested-1,ic:ic+nbColBlockTested-1);
        
        % initialization to look for min Cost: SAE (sum of absolute
        % errors), square error
        CostMin=inf;
        for shiftCol=-s:s,
            for shiftRow=-s:s,

                % restriction to full blocks inside the frame
                if (ir-shiftRow>=1 && ir-shiftRow+nbRowBlockTested-1<= nbRow &&  ic-shiftCol>=1 && ic-shiftCol+nbColBlockTested-1<=nbCol),
                    referenceBlock=referenceImage(ir-shiftRow:ir-shiftRow+nbRowBlockTested-1,ic-shiftCol:ic-shiftCol+nbColBlockTested-1);
                    % -shiftRow to keep some 
                    % Cost=sum(abs(referenceBlock(:)-currentBlock(:))); % SAE
                    Cost=sum((referenceBlock(:)-currentBlock(:)).^2); % Square error
                    %fprintf('  %d/%d: %d\n',shiftRow,shiftCol,Cost); % debug
                    if (Cost<CostMin)
                        CostMin=Cost;
                        predictedBlock=referenceBlock;
                        shiftColMin=shiftCol;
                        shiftRowMin=shiftRow;
                     end % if
                end % if
            end % loop shiftRow
        end % loop shiftCol
        %
        predictedImage(ir:ir+nbRowBlockTested-1,ic:ic+nbColBlockTested-1) = predictedBlock;
        MotionVectorPosition=[MotionVectorPosition ; ir+round(nbRowBlockTested/2) , ic+round(nbColBlockTested/2)];% does not need to be transmitted. Can be recomputed at decoder
        MotionVector=[MotionVector ; shiftRowMin, shiftColMin];
        %fprintf(' MV: %d %d\n',shiftRowMin,shiftColMin); % debug
        
    end % loop ic on column
end % loop ir on row
toc
%% Rate Distortion evaluation

% Rate: Entropy coding of motion vectors

mvjoint=MotionVector(:,1) + 100 * MotionVector(:,2); % 100 is >> range (MotionVector(:,1))
% thus MotionVector(:,1) and MotionVector(:,2) can be retrieved without any ambiguity
% from mvjoint
mm=min(mvjoint);
MM=max(mvjoint);
p=hist(mvjoint,mm:MM);
p=p(find(p)); % to avoid to compute log2(0) in the entropy
p = p/sum(p); % normalization of p
H = sum (- p.*log2(p)) ;
JointEntropy= H;

%MarginalEntropy
p=hist(MotionVector(:,1));
p=p(find(p)); % to avoid to compute log2(0) in the entropy
p = p/sum(p); % normalization of p
HR = sum (- p.*log2(p)) ;
p=hist(MotionVector(:,2));
p=p(find(p)); % to avoid to compute log2(0) in the entropy
p = p/sum(p); % normalization of p
HC = sum (- p.*log2(p)) ;
MarginalEntropy= HR + HC;

fprintf('cost:\n joint %1.5f per motion vector \n marginal %1.5f per motion vector\n',JointEntropy,MarginalEntropy)
Rate= round(JointEntropy * size(MotionVectorPosition,1));

% Distortion
diff=predictedImage-currentImage;
err=sum(sum(diff.*diff));
err=err/numel(diff);     
PSNR=10*log10(255^2/err);

fprintf('Rate %d bits - Distortion %1.2f dB\n',Rate,PSNR);


%% Show images after motion compensation
figure(4)
image(predictedImage); colormap(gray(256)); axis image; axis off; title('Motion compensated image');

figure(5)
image(128+currentImage-predictedImage); colormap(gray(256)); axis image; axis off; title('Motion compensated error');

figure(6)
%image(currentImage); colormap(gray(256)); axis image; %axis off; 
image(predictedImage); colormap(gray(256)); axis image; %axis off; 
title(sprintf('Frame: %d, Cost: %d bits, PSNR: %1.2f dB',currentFrameNumber,Rate,PSNR));
hold on

% draw block boundaries
for ir=nbRowBlock:nbRowBlock:nbRow-1, % no block boundaries for the image boundary
    line([1,nbCol],[ir-.5,ir-.5]);
end
for ic=nbColBlock:nbColBlock:nbCol-1,
    line([ic-.5,ic-.5],[1,nbCol]);
end

% draw Motion vectors
% vectors starts in the centre of the current block
%quiver(MotionVectorPosition(:,2), MotionVectorPosition(:,1),MotionVector(:,2),MotionVector(:,1),0) % ,0 to avoi automatic scaling of the vectors
% vectors starts in the centre of the block taken from the reference
quiver(MotionVectorPosition(:,2)-MotionVector(:,2), MotionVectorPosition(:,1)-MotionVector(:,1),MotionVector(:,2),MotionVector(:,1),0)


hold off



